@@ -6,11 +6,11 @@
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 {-|
 Module      : HsLua.Module.DocLayout
-Copyright   : © 2020-2024 Albert Krewinkel
+Copyright   : © 2020-2026 Albert Krewinkel
 License     : MIT
 Maintainer  : Albert Krewinkel <albert+hslua@zeitkraut.de>
 
-Provides a Lua module which wraps @'Text.DocLayout'@. The @Doc'
+Provides a Lua module which wraps @'Text.DocLayout'@. The @'Doc'@
 type is specialized to @'Text'@.
 
 This module defines orphan instances for @Doc Text@.
@@ -63,6 +63,14 @@ module HsLua.Module.DocLayout (
   , real_length
   , update_column
 
+  -- * Styling
+  , bold
+  , italic
+  , underlined
+  , strikeout
+  , fg
+  , bg
+
   -- * Marshaling
   , peekDoc
   , pushDoc
@@ -76,6 +84,7 @@ import HsLua as Lua hiding (concat)
 import Text.DocLayout (Doc, (<+>), ($$), ($+$))
 
 import qualified Data.Text as T
+import qualified Data.Text.Lazy as TL
 import qualified Text.DocLayout as Doc
 
 --
@@ -88,14 +97,11 @@ description = "Plain-text document layouting."
 
 -- | Self-documenting module.
 documentedModule :: LuaError e => Module e
-documentedModule = Module
-  { moduleName = "doclayout"
-  , moduleFields = fields
-  , moduleDescription = description
-  , moduleFunctions = functions
-  , moduleOperations = []
-  , moduleTypeInitializers = [initType typeDoc]
-  }
+documentedModule = defmodule "doclayout"
+  `withFields` fields
+  `withDescription` description
+  `withFunctions` functions
+  `associateType` typeDoc
 
 --
 -- Fields
@@ -112,41 +118,33 @@ fields =
 
 -- | Wrapped and documented 'Doc.blankline' value.
 blankline :: forall e. LuaError e => Field e
-blankline = Field
-  { fieldName = "blankline"
-  , fieldDescription = "Inserts a blank line unless one exists already."
-  , fieldType = udTypeSpec @e typeDoc
-  , fieldPushValue = pushDoc Doc.blankline
-  }
+blankline = deffield "blankline"
+  `withDescription` "Inserts a blank line unless one exists already."
+  `withType` udTypeSpec @e typeDoc
+  `withValue` pushDoc Doc.blankline
 
 -- | Wrapped and documented 'Doc.cr' value.
 cr :: forall e. LuaError e => Field e
-cr = Field
-  { fieldName = "cr"
-  , fieldDescription = "A carriage return. Does nothing if we're at " <>
-                       "the beginning of a line; " <>
-                       "otherwise inserts a newline."
-  , fieldType = udTypeSpec @e typeDoc
-  , fieldPushValue = pushDoc Doc.cr
-  }
+cr = deffield "cr"
+  `withDescription`
+    "A carriage return. Does nothing if we're at " <>
+    "the beginning of a line; otherwise inserts a newline."
+  `withType` udTypeSpec @e typeDoc
+  `withValue` pushDoc Doc.cr
 
 -- | Wrapped and documented 'Doc.empty' value.
 empty :: forall e. LuaError e => Field e
-empty = Field
-  { fieldName = "empty"
-  , fieldDescription = "The empty document."
-  , fieldType = udTypeSpec @e typeDoc
-  , fieldPushValue = pushDoc Doc.empty
-  }
+empty = deffield "empty"
+  `withDescription` "The empty document."
+  `withType` udTypeSpec @e typeDoc
+  `withValue` pushDoc Doc.empty
 
 -- | Wrapped and documented 'Doc.space' value.
 space :: forall e. LuaError e => Field e
-space = Field
-  { fieldName = "space"
-  , fieldDescription = "A breaking (reflowable) space."
-  , fieldType = udTypeSpec @e typeDoc
-  , fieldPushValue = pushDoc Doc.space
-  }
+space = deffield "space"
+  `withDescription` "A breaking (reflowable) space."
+  `withType` udTypeSpec @e typeDoc
+  `withValue` pushDoc Doc.space
 
 --
 -- Functions
@@ -187,6 +185,13 @@ functions =
   , offset
   , real_length
   , update_column
+    -- styling
+  , bold
+  , italic
+  , underlined
+  , strikeout
+  , fg
+  , bg
   ]
 
 typeDoc :: LuaError e => DocumentedType e (Doc Text)
@@ -243,15 +248,31 @@ typeDoc = deftype "Doc"
 -- line length parameter is omitted or nil.
 render :: LuaError e => DocumentedFunction e
 render = defun "render"
-  ### liftPure2 (flip Doc.render)
+  ### (\doc mbcolwidth useAnsi ->
+          if useAnsi == Just True
+          then pure $ TL.toStrict $ Doc.renderANSI mbcolwidth doc
+          else pure $ Doc.renderPlain mbcolwidth doc)
   <#> docParam "doc"
-  <#> opt (integralParam "colwidth" "planned maximum line length")
-  =#> functionResult pushText "Doc" "rendered doc"
+  <#> opt (integralParam "colwidth" $
+           "Maximum number of characters per line.\n" <>
+           "A value of `nil`, the default, means that the text " <>
+           "is not reflown.")
+  <#> opt (parameter peekRenderStyle "string" "style" $
+           "Whether to generate plain text or ANSI terminal output.\n" <>
+           "Must be either `'plain'` or `'ansi'`.\n" <>
+           "Defaults to `'plain'`.")
+  =#> functionResult pushText "string" "rendered doc"
   #? T.unlines
      [ "Render a [[Doc]]. The text is reflowed on breakable spaces to"
      , "match the given line length. Text is not reflowed if the line"
      , "line length parameter is omitted or nil."
      ]
+ where
+   peekRenderStyle idx = peekByteString idx >>= \case
+     "ansi"  -> pure True
+     "ANSI"  -> pure True
+     "plain" -> pure False
+     style   -> failPeek $ "Unknown rendering style: " <> style
 
 --
 -- Querying
@@ -312,7 +333,7 @@ real_length = defun "real_length"
   <#> textParam "str" "UTF-8 string to measure"
   =#> integralResult "text length"
   #? ("Returns the real length of a string in a monospace font: " <>
-      "0 for a combining chaeracter, 1 for a regular character, " <>
+      "0 for a combining character, 1 for a regular character, " <>
       "2 for an East Asian wide character.")
 
 --
@@ -354,11 +375,11 @@ braces :: LuaError e => DocumentedFunction e
 braces = defun "braces"
   ### liftPure Doc.braces
   <#> docParam "doc"
-  =#> docResult "doc enclosed by {}."
+  =#> docResult "`doc` enclosed by {}."
   #? "Puts the `doc` in curly braces."
 
 -- | Puts a @'Doc'@ in square brackets.
-brackets :: LuaError e => DocumentedFunction e -- Doc Text -> Lua (Doc Text)
+brackets :: LuaError e => DocumentedFunction e
 brackets = defun "brackets"
   ### liftPure Doc.brackets
   <#> docParam "doc"
@@ -371,7 +392,7 @@ cblock = defun "cblock"
   ### liftPure2 (flip Doc.cblock)
   <#> docParam "doc"
   <#> integralParam "width" "block width in chars"
-  =#> docResult ("doc, aligned centered in a block with max" <>
+  =#> docResult ("doc, aligned centered in a block with max " <>
                  "`width` chars per line.")
   #? ("Creates a block with the given width and content, " <>
       "aligned centered.")
@@ -511,7 +532,7 @@ rblock = defun "rblock"
   ### liftPure2 (flip Doc.rblock)
   <#> docParam "doc"
   <#> integralParam "width" "block width in chars"
-  =#> docResult ("doc, right aligned in a block with max" <>
+  =#> docResult ("doc, right aligned in a block with max " <>
                  "`width` chars per line.")
   #? ("Creates a block with the given width and content, " <>
       "aligned to the right.")
@@ -528,6 +549,54 @@ vfill = defun "vfill"
       "expands to the height of the box.  Strings cycle through the " <>
       "list provided.")
 
+--
+-- Styling
+--
+
+bold :: LuaError e => DocumentedFunction e
+bold = defun "bold"
+  ### liftPure Doc.bold
+  <#> docParam "doc"
+  =#> docResult "bolded Doc"
+  #? "Puts a [[Doc]] in boldface."
+
+italic :: LuaError e => DocumentedFunction e
+italic = defun "italic"
+  ### liftPure Doc.italic
+  <#> docParam "doc"
+  =#> docResult "styled Doc"
+  #? "Puts a [[Doc]] in italics."
+
+underlined :: LuaError e => DocumentedFunction e
+underlined = defun "underlined"
+  ### liftPure Doc.underlined
+  <#> docParam "doc"
+  =#> docResult "styled Doc"
+  #? "Underlines a [[Doc]]."
+
+strikeout :: LuaError e => DocumentedFunction e
+strikeout = defun "strikeout"
+  ### liftPure Doc.strikeout
+  <#> docParam "doc"
+  =#> docResult "styled Doc"
+  #? "Puts a line through the [[Doc]]."
+
+fg :: LuaError e => DocumentedFunction e
+fg = defun "fg"
+  ### liftPure2 (flip Doc.fg)
+  <#> docParam "doc"
+  <#> colorParam
+  =#> docResult "styled Doc"
+  #? "Set the foreground color."
+
+bg :: LuaError e => DocumentedFunction e
+bg = defun "bg"
+  ### liftPure2 (flip Doc.bg)
+  <#> docParam "doc"
+  <#> colorParam
+  =#> docResult "styled Doc"
+  #? "Set the background color."
+
 --
 -- Marshaling
 --
@@ -561,6 +630,23 @@ instance Pushable (Doc Text) where
 docParam :: LuaError e => Text -> Parameter e (Doc Text)
 docParam name = parameter peekDoc "Doc" name "document"
 
+-- | @Color@ function parameter
+colorParam :: Parameter e Doc.Color
+colorParam = parameter peekColor "string" "color"
+  ("One of 'black', 'red', 'green', 'yellow', 'blue', 'magenta' " <>
+   "'cyan', or 'white'.")
+  where
+    peekColor idx = peekByteString idx >>= \case
+      "black"   -> pure Doc.black
+      "red"     -> pure Doc.red
+      "green"   -> pure Doc.green
+      "yellow"  -> pure Doc.yellow
+      "blue"    -> pure Doc.blue
+      "magenta" -> pure Doc.magenta
+      "cyan"    -> pure Doc.cyan
+      "white"   -> pure Doc.white
+      color -> failPeek $ "Unknown color: " <> color
+
 --
 -- Results
 --
